#include "config.h"

#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>

#include <pthread.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <grp.h>

#ifdef HAVE_SYS_SYSCALL_H
#include <sys/syscall.h>
#endif
#ifdef HAVE_SYSCALL_H
#include <syscall.h>
#endif

#define NUM_THREADS 10

#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))

struct parm {
	int id;
	int ready;
};

pthread_mutex_t msg_mutex = PTHREAD_MUTEX_INITIALIZER;

static void *syscall_setreuid(void *arg)
{
	long int rc;
	uid_t u;
	uid_t ru;

	(void) arg; /* unused */

	/* This load can help with revealing race conditions. */
	for (ru = 0; ru < 2048; ++ru) {
		rc = syscall(SYS_setreuid, -1, ru);
		assert_int_equal(rc, 0);
		u = geteuid();
		assert_int_equal(u, ru);
	}

	return NULL;
}

static void test_syscall_setreuid(void **state)
{
	pthread_attr_t pthread_custom_attr;
	pthread_t threads[NUM_THREADS];
	uid_t u;
	int rc;
	int i;

	(void) state; /* unused */

	rc = seteuid(666);
	assert_int_equal(rc, 0);

	pthread_attr_init(&pthread_custom_attr);

	for (i = 0; i < NUM_THREADS; i++) {
		pthread_create(&threads[i],
			       &pthread_custom_attr,
			       syscall_setreuid,
			       NULL);
	}

	for (i = 0; i < NUM_THREADS; i++) {
		pthread_join(threads[i], NULL);
	}

	u = geteuid();
	assert_int_equal(u, 666);

	pthread_attr_destroy(&pthread_custom_attr);
}

static void *sync_setreuid(void *arg)
{
	struct parm *p = (struct parm *)arg;
	uid_t u;

	syscall_setreuid(arg);

	p->ready = 1;

	pthread_mutex_lock(&msg_mutex);

	u = geteuid();
	assert_int_equal(u, 42);

	pthread_mutex_unlock(&msg_mutex);

	return NULL;
}

static void test_sync_setreuid(void **state)
{
	pthread_attr_t pthread_custom_attr;
	pthread_t threads[NUM_THREADS];
	struct parm *p;
	int rc;
	int i;

	(void) state; /* unused */

	pthread_attr_init(&pthread_custom_attr);

	p = malloc(NUM_THREADS * sizeof(struct parm));
	assert_non_null(p);

	pthread_mutex_lock(&msg_mutex);

	for (i = 0; i < NUM_THREADS; i++) {
		p[i].id = i;
		p[i].ready = 0;

		pthread_create(&threads[i],
			       &pthread_custom_attr,
			       sync_setreuid,
			       (void *)&p[i]);
	}

	/* wait for the reads to set euid to 0 */
	for (i = 0; i < NUM_THREADS; i++) {
		while (p[i].ready != 1) {
			sleep(1);
		}
	}

	rc = setreuid(-1, 42);
	assert_int_equal(rc, 0);

	pthread_mutex_unlock(&msg_mutex);

	for (i = 0; i < NUM_THREADS; i++) {
		pthread_join(threads[i], NULL);
	}

	pthread_attr_destroy(&pthread_custom_attr);
	free(p);
}

static void *uwrap_setgid_syscall(void *arg)
{
	int rc;
	gid_t g;
	(void) arg; /* unused */

	rc = syscall(SYS_setgid, 999);
	assert_int_equal(rc, 0);

	g = getgid();
	assert_int_equal(g, 999);

	return NULL;
}

static void test_sync_setgid_syscall(void **state)
{
	pthread_attr_t pthread_custom_attr;
	pthread_t threads[NUM_THREADS];
	gid_t go, gn;
	int i;

	(void) state; /* unused */

	pthread_attr_init(&pthread_custom_attr);

	go = getgid();

	for (i = 0; i < NUM_THREADS; i++) {
		pthread_create(&threads[i],
			       &pthread_custom_attr,
			       uwrap_setgid_syscall,
			       NULL);
	}

	/* wait for threaads */
	for (i = 0; i < NUM_THREADS; i++) {
		pthread_join(threads[i], NULL);
	}

	gn = getgid();
	assert_int_equal(gn, go);

	pthread_attr_destroy(&pthread_custom_attr);
}

static void *uwrap_setgid(void *arg)
{
	int rc;
	(void) arg; /* unused */

	rc = setgid(999);
	assert_int_equal(rc, 0);

	return NULL;
}

static void *uwrap_getgid(void *arg)
{
	gid_t g;
	(void) arg; /* unused */

	g = getgid();
	assert_int_equal(g, 1999);

	return NULL;
}

static void test_sync_setgid(void **state)
{
	pthread_attr_t pthread_custom_attr;
	pthread_t threads[NUM_THREADS];
	gid_t g;
	int i;
	int rc;

	(void) state; /* unused */

	pthread_attr_init(&pthread_custom_attr);

	for (i = 0; i < NUM_THREADS; i++) {
		pthread_create(&threads[i],
			       &pthread_custom_attr,
			       uwrap_setgid,
			       NULL);
	}

	/* wait for threaads */
	for (i = 0; i < NUM_THREADS; i++) {
		pthread_join(threads[i], NULL);
	}

	g = getgid();
	assert_int_equal(g, 999);

	rc = setgid(1999);
	assert_int_equal(rc, 0);

	pthread_create(&threads[0],
		       &pthread_custom_attr,
		       uwrap_getgid,
		       NULL);

	pthread_join(threads[0], NULL);

	g = getgid();
	assert_int_equal(g, 1999);

	pthread_attr_destroy(&pthread_custom_attr);
}


static pthread_mutex_t wait_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t sleep_mutex = PTHREAD_MUTEX_INITIALIZER;
static void *uwrap_getuid_sync(void *arg)
{
	uid_t u;

	(void) arg; /* unused */

	pthread_mutex_unlock(&sleep_mutex);
	pthread_mutex_lock(&wait_mutex);

	u = getuid();
	assert_int_equal(u, 888);

	return NULL;
}

static void *uwrap_setuid_sync(void *arg)
{
	int rc;

	(void) arg; /* unused */

	rc = setuid(888);
	assert_int_equal(rc, 0);

	return NULL;
}

static void test_real_sync_setuid(void **state)
{
	pthread_t threads[2];
	uid_t u;
	int rc;

	(void) state; /* unused */

	rc = setuid(222);
	assert_int_equal(rc, 0);

	pthread_mutex_lock(&wait_mutex);
	pthread_mutex_lock(&sleep_mutex);

	/* Create thread which will wait for change. */
	pthread_create(&threads[0],
		       NULL,
		       uwrap_getuid_sync,
		       NULL);

	pthread_mutex_lock(&sleep_mutex);

	pthread_create(&threads[1],
		       NULL,
		       uwrap_setuid_sync,
		       NULL);

	pthread_join(threads[1], NULL);

	pthread_mutex_unlock(&wait_mutex);

	pthread_join(threads[0], NULL);

	u = getuid();
	assert_int_equal(u, 888);
}

static void *uwrap_setgroups(void *arg)
{
	gid_t glist[] = { 100, 200, 300, 400, 500 };
	int rc;

	(void) arg; /* unused */

	rc = setgroups(ARRAY_SIZE(glist), glist);
	assert_int_equal(rc, 0);

	return NULL;
}

static void *uwrap_getgroups(void *arg)
{
	gid_t glist[] = { 100, 200, 300, 400, 500 };
	gid_t rlist[16];
	int rc;

	(void) arg; /* unused */

	rc = getgroups(ARRAY_SIZE(rlist), rlist);
	assert_int_equal(rc, 5);

	assert_memory_equal(glist, rlist, sizeof(glist));

	return NULL;
}

static void test_sync_setgroups(void **state)
{
	gid_t glist[] = { 100, 200, 300, 400, 500 };
	pthread_t threads[NUM_THREADS];
	gid_t rlist[16];
	int rc;
	int i;

	(void) state; /* unused */

	for (i = 0; i < NUM_THREADS; i++) {
		pthread_create(&threads[i],
			       NULL,
			       uwrap_setgroups,
			       NULL);
	}

	/* wait for threaads */
	for (i = 0; i < NUM_THREADS; i++) {
		pthread_join(threads[i], NULL);
	}

	rc = getgroups(ARRAY_SIZE(rlist), rlist);
	assert_int_equal(rc, 5);

	assert_memory_equal(glist, rlist, sizeof(glist));

	pthread_create(&threads[0],
		       NULL,
		       uwrap_getgroups,
		       NULL);

	pthread_join(threads[0], NULL);

}

static void *uwrap_create_thread_setgid(void *arg)
{
	pthread_t thread;
	signed int tmp;
	gid_t g, g_a;
	int rc;

	tmp = *((signed int *) arg);

	if (tmp >= 0) {
		g_a = (gid_t) *((unsigned int *) arg);

		rc = syscall(SYS_setgid, g_a);
		assert_int_equal(rc, 0);

		g = getgid();
		assert_int_equal(g, g_a);

		tmp *= -1;
		rc = pthread_create(&thread,
				NULL,
				&uwrap_create_thread_setgid,
				(void *) &tmp);
		assert_int_equal(rc, 0);

		pthread_join(thread, NULL);

		g = getgid();
		assert_int_equal(g, g_a);
	} else {
		tmp *= -1;
		g_a = (gid_t) tmp;
		g = getgid();
		assert_int_equal(g, g_a);
	}

	pthread_exit(NULL);
}

static void test_thread_create_thread_setgid(void **state)
{
	pthread_t thread;
	signed int tmp = 666;
	gid_t g;
	int rc;

	(void) state; /* unused */

	rc = setgid(555);
	assert_int_equal(rc, 0);
	g = getgid();
	assert_int_equal(g, 555);

	pthread_create(&thread,
			NULL,
			&uwrap_create_thread_setgid,
			&tmp);
	pthread_join(thread, NULL);
}

int main(void) {
	int rc;

	const UnitTest tests[] = {
		unit_test(test_syscall_setreuid),
		unit_test(test_sync_setreuid),
		unit_test(test_sync_setgid),
		unit_test(test_sync_setgid_syscall),
		unit_test(test_real_sync_setuid),
		unit_test(test_sync_setgroups),
		unit_test(test_thread_create_thread_setgid),
	};

	rc = run_tests(tests);

	return rc;
}
